//----------------------------------------------------------------------------------
// aviutl  GPUJX^tB^p ̂ق̊֐
// Copyright (c) 2008  N099
//----------------------------------------------------------------------------------
#include <windows.h>
#include <emmintrin.h>
#include <intrin.h>
#include <math.h>

#include "filter.h"
#include "util.h"

//---------------------------------------------------------------------
// internal functions
//---------------------------------------------------------------------
static const int SIMD_MMX  = 1;
static const int SIMD_SSE  = 2;
static const int SIMD_SSE2 = 4;
static int check_simd();

inline void yc_f32(void*, const PIXEL_YC*, unsigned int);
inline void yc_f32_sse(void*, const PIXEL_YC*, unsigned int);
inline void yc_f32_sse2(void*, const PIXEL_YC*, unsigned int);
inline void yc_f16(void*, const PIXEL_YC*, unsigned int);
inline void yc_f16_sse(void*, const PIXEL_YC*, unsigned int);
inline void yc_f16_sse2(void*, const PIXEL_YC*, unsigned int);
inline void yc_i16_mmx(void*, const PIXEL_YC*, unsigned int);
inline void yc_i16_sse2(void*, const PIXEL_YC*, unsigned int);
inline void yc_b16_mmx(void*, const PIXEL_YC*, unsigned int);
inline void yc_b16_sse2(void*, const PIXEL_YC*, unsigned int);

inline void f32_yc(PIXEL_YC*, const void*, unsigned int);
inline void f32_yc_sse(PIXEL_YC*, const void*, unsigned int);
inline void f32_yc_sse2(PIXEL_YC*, const void*, unsigned int);
inline void f16_yc(PIXEL_YC*, const void*, unsigned int);
inline void f16_yc_sse(PIXEL_YC*, const void*, unsigned int);
inline void f16_yc_sse2(PIXEL_YC*, const void*, unsigned int);
inline void i16_yc_mmx(PIXEL_YC*, const void*, unsigned int);
inline void i16_yc_sse2(PIXEL_YC*, const void*, unsigned int);
inline void b16_yc_mmx(PIXEL_YC*, const void*, unsigned int);
inline void b16_yc_sse2(PIXEL_YC*, const void*, unsigned int);

inline short f32_f16(float f);
inline float f16_f32(short s);

//---------------------------------------------------------------------
//PIXEL_YC̕ϊ(float)
// Y (0, 4096) <-> (0.0, 1.0)
// C (-2048, 2048) <-> (-0.5, 0.5)

//t[̃TCY
// PIXEL_YC
//        w
//  ==============----- 
//  *            *    |
// h*            *    |
//  *            *    |max_h
//  *============*    |
//  |       max_w     |
//  -------------------
//
// Texture
//        w
//  ==============------- 
//  *            *      |
// h*            *      |
//  *            *      |
//  *============*      |sh
//  |                   |
//  |        sw         |
//  ---------------------
//
// float -> short ̊ۂ߂ /QIfist IvV𗘗p
// SSE𗘗p邱Ƃł؎̂Ăł(CVTPS2PI)
//---------------------------------------------------------------------

//---------------------------------------------------------------------
// union
//---------------------------------------------------------------------
union u32_m128i { unsigned __int32 u[4]; __m128i xi; };
union u64_m128i { __m64 m[2]; __m128i xi; };

//---------------------------------------------------------------------
// const
//---------------------------------------------------------------------
static const __int32 f16_scl = 0x3dfff000; //2^(12+112) - 2^12
static const __int32 f16_rcl = 0x3e000000; //2^(12+112)
static const __m64 b16_ofs  = { 0x8000800080008000UI64 };
static const __m64 i16_msk  = { 0x0000ffffffffffffUI64 };
static const __m64 f16_scl2 = { 0x3dfff0003dfff000UI64 };
static const __m64 f16_rcl2 = { 0x3e0000003e000000UI64 };
static const __m128 f32_scl = { 1.0f/4096.0f, 1.0f/4096.0f, 1.0f/4096.0f, 0.0f };
static const __m128 f32_rcl = { 4096.0f, 4096.0f, 4096.0f, 0.0f };
static const u32_m128i f16_sgn  = { 1U << 31, 1U << 31, 1U << 31, 1U << 31 };
static const u64_m128i f16_scl3 = { f16_scl2, f16_scl2 };
static const u64_m128i f16_rcl3 = { f16_rcl2, f16_rcl2 };
static const u64_m128i b16_ofs2 = { b16_ofs, b16_ofs };
static const u64_m128i i16_msk2 = { i16_msk, i16_msk };
static const u64_m128i f16_msk2 = { 0xffffffffffffffffUI64, 0xffffffffUI64 };

//---------------------------------------------------------------------
// float <-> half w/ scaling
//---------------------------------------------------------------------
short f32_f16(float f)
{
	short tmp;
	union { float t; int c; };
	t = f;

	if(c == 0) return 0; //+0.0f
	
	c -= f16_scl; //rounding - exponent bias / 4096.0f
	tmp = (short)((c >> 16) & 0x8000); //sign
	tmp |= (short)((c >> 13) & 0x7fff);

	return tmp;
}

float f16_f32(short s)
{
	int tmp;
	union { float t; int c; };
	c = s & 0x7fff;

	if(c == 0) return 0.0f; //+-0.0f

	tmp = s & 0x8000; //sign
	c = (c << 13) | (tmp << 16);
	c += f16_rcl; //exponent bias * 4096.0f

	return t;
}

//---------------------------------------------------------------------
// 1C
//---------------------------------------------------------------------

//ϊRs[(line) PIXEL_YC -> D3DFMT_A32B32G32R32F
void yc_f32(void *dst, const PIXEL_YC *src, unsigned int n)
{
	const float r = 1.0f/4096.0f;
	float *fdst = static_cast<float*>(dst);

	for(unsigned int i=0; i<n; ++i){
		*fdst++ = r * src->y;
		*fdst++ = r * src->cb;
		*fdst++ = r * src->cr;
		*fdst++ = 0.0f;
		src++;
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_A16B16G16R16F
void yc_f16(void *dst, const PIXEL_YC *src, unsigned int n)
{
	short *sdst = static_cast<short*>(dst);

	for(unsigned int i=0; i<n; ++i){
		*sdst++ = f32_f16(src->y);
		*sdst++ = f32_f16(src->cb);
		*sdst++ = f32_f16(src->cr);
		*sdst++ = 0;
		src++;
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_Q16W16V16U16 MMXp
void yc_i16_mmx(void *dst, const PIXEL_YC *src, unsigned int n)
{
	__m64 *mdst = static_cast<__m64*>(dst);

	for(unsigned int i=0; i<n; ++i){
		*mdst++ = _mm_and_si64(*(__m64*)(src++), i16_msk);
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_A16B16G16R16 MMXp
void yc_b16_mmx(void *dst, const PIXEL_YC *src, unsigned int n)
{
	__m64 *mdst = static_cast<__m64*>(dst);

	for(unsigned int i=0; i<n; ++i){
		*mdst++ = _mm_and_si64(_mm_add_pi16(*(__m64*)src++, b16_ofs), i16_msk);
	}
}

//ϊRs[ D3DFMT_A32B32G32R32F -> PIXEL_YC
void f32_yc(PIXEL_YC *dst, const void *src, unsigned int n)
{
	const float *fsrc = static_cast<const float*>(src);

	for(unsigned int i=0; i<n; ++i){
		dst->y  = (short)(floor(0.5f + 4096.0f * *fsrc++));
		dst->cb = (short)(floor(0.5f + 4096.0f * *fsrc++));
		dst->cr = (short)(floor(0.5f + 4096.0f * *fsrc++));
		fsrc++;
		dst++;
	}
}

//ϊRs[ D3DFMT_A16B16G16R16F -> PIXEL_YC
void f16_yc(PIXEL_YC *dst, const void *src, unsigned int n)
{
	const short *ssrc = static_cast<const short*>(src);

	for(unsigned int i=0; i<n; ++i){
		dst->y  = (short)(floor(0.5f + f16_f32(*ssrc++)));
		dst->cb = (short)(floor(0.5f + f16_f32(*ssrc++)));
		dst->cr = (short)(floor(0.5f + f16_f32(*ssrc++)));
		ssrc++;
		dst++;
	}
}

//ϊRs[ D3DFMT_Q16W16V16U16 -> PIXEL_YC MMXp
void i16_yc_mmx(PIXEL_YC *dst, const void *src, unsigned int n)
{
	const __m64 *msrc = static_cast<const __m64*>(src);

	for(unsigned int i=0; i<n; ++i){
		*(__m64*)dst++ = *msrc++;
	}
}

//ϊRs[ D3DFMT_A16B16G16R16 -> PIXEL_YC MMXp
void b16_yc_mmx(PIXEL_YC *dst, const void *src, unsigned int n)
{
	const __m64 *msrc = static_cast<const __m64*>(src);

	for(unsigned int i=0; i<n; ++i){
		*(__m64*)dst++ = _mm_sub_pi16(*msrc++, b16_ofs);
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_A32B32G32R32F SSEp
void yc_f32_sse(void *dst, const PIXEL_YC *src, unsigned int n)
{
	float *fdst = static_cast<float*>(dst);

	for(unsigned int i=0; i<n; ++i){
		_mm_storeu_ps(fdst, _mm_mul_ps(
			_mm_cvtpi16_ps(*(const __m64*)src++), f32_scl));
		fdst += 4;
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_A16B16G16R16F SSEp
void yc_f16_sse(void *dst, const PIXEL_YC *src, unsigned int n)
{
	__m64 mi,mj,ms,mz;
	__m64 *mdst = static_cast<__m64*>(dst);

	for(unsigned int i=0; i<n; ++i){
		mi = *(const __m64*)src++;
		mz = _mm_cmpeq_pi16(_mm_setzero_si64(), mi);
		ms = _mm_and_si64(b16_ofs, mi);
		mz = _mm_andnot_si64(mz, i16_msk);

		mj = _mm_srai_pi32(_mm_unpackhi_pi16(_mm_setzero_si64(), mi), 16);
		mi = _mm_srai_pi32(_mm_unpacklo_pi16(mi, mi), 16);
		_mm_storel_pi(&mi, _mm_cvtpi32_ps(_mm_setzero_ps(), mi));
		_mm_storel_pi(&mj, _mm_cvtpi32_ps(_mm_setzero_ps(), mj));

		mi = _mm_sub_pi32(mi, f16_scl2);
		mj = _mm_sub_pi32(mj, f16_scl2);
		mi = _mm_srai_pi32(_mm_slli_pi32(mi, 3), 16);
		mj = _mm_srai_pi32(_mm_slli_pi32(mj, 3), 16);
		mi = _mm_packs_pi32(mi, mj);

		mi = _mm_and_si64(_mm_or_si64(mi, ms), mz);
		_mm_stream_pi(mdst++, mi);
	}
}

//ϊRs[ D3DFMT_A32B32G32R32F -> PIXEL_YC SSEp
void f32_yc_sse(PIXEL_YC *dst, const void *src, unsigned int n)
{
	const float *fsrc = static_cast<const float*>(src);

	for(unsigned int i=0; i<n; ++i){
		*(__m64*)dst++ = _mm_cvtps_pi16(
			_mm_mul_ps(_mm_loadu_ps(fsrc), f32_rcl));
		fsrc += 4;
	}
}

//ϊRs[ D3DFMT_A16B16G16R16F -> PIXEL_YC SSEp
void f16_yc_sse(PIXEL_YC *dst, const void *src, unsigned int n)
{
	__m64 mi,mj,ms,mt,my,mz;
	const __m64 *msrc = static_cast<const __m64*>(src);

	for(unsigned int i=0; i<n; ++i){
		mj = *msrc++;
		mi = _mm_unpacklo_pi16(_mm_setzero_si64(), mj);
		mj = _mm_unpackhi_pi16(_mm_setzero_si64(), mj);
		
		ms = _mm_and_si64(b16_ofs, mi);
		mt = _mm_and_si64(b16_ofs, mj);
		mi = _mm_andnot_si64(b16_ofs, mi);
		mj = _mm_andnot_si64(b16_ofs, mj);
		my = _mm_cmpeq_pi32(_mm_setzero_si64(), mi);
		mz = _mm_cmpeq_pi32(_mm_setzero_si64(), mj);
		
		mi = _mm_srli_pi32(mi, 3);
		mj = _mm_srli_pi32(mj, 3);
		mi = _mm_add_pi32(mi, f16_rcl2);
		mj = _mm_add_pi32(mj, f16_rcl2);
		mi = _mm_andnot_si64(my, _mm_or_si64(mi, ms));
		mj = _mm_andnot_si64(mz, _mm_or_si64(mj, mt));

		mi = _mm_cvtps_pi32(_mm_loadl_pi(_mm_setzero_ps(), &mi));
		mj = _mm_cvtps_pi32(_mm_loadl_pi(_mm_setzero_ps(), &mj));
		*(__m64*)dst++ = _mm_packs_pi32(mi, mj);
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_A32B32G32R32F SSE2p
void yc_f32_sse2(void *dst, const PIXEL_YC *src, unsigned int n)
{
	__m128i xi;
	float *fdst = static_cast<float*>(dst);

	for(unsigned int i=0; i<n; ++i){
		xi = _mm_loadl_epi64((const __m128i*)src++);
		xi = _mm_srai_epi32(_mm_unpacklo_epi16(xi, xi), 16);
		_mm_storeu_ps(fdst, _mm_mul_ps(_mm_cvtepi32_ps(xi), f32_scl));
		fdst += 4;
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_A16B16G16R16F SSE2p
void yc_f16_sse2(void *dst, const PIXEL_YC *src, unsigned int n)
{
	__m128i xi,xs,xz;
	short *sdst = static_cast<short*>(dst);

	for(unsigned int i=0; i<n; ++i){
		xi = _mm_loadl_epi64((const __m128i*)src++);
		xi = _mm_srai_epi32(_mm_unpacklo_epi16(xi, xi), 16);
		xi = _mm_castps_si128(_mm_cvtepi32_ps(xi));
		xz = _mm_cmpeq_epi32(_mm_setzero_si128(), xi);
		xz = _mm_andnot_si128(xz, f16_msk2.xi);
		xs = _mm_and_si128(xi, f16_sgn.xi);
		xi = _mm_sub_epi32(xi, f16_scl3.xi);
		xi = _mm_slli_epi32(xi, 3);
		xi = _mm_or_si128(xi, xs);
		xi = _mm_srai_epi32(_mm_and_si128(xi, xz), 16);
		_mm_storel_epi64((__m128i*)sdst, _mm_packs_epi32(xi, xi));
		sdst += 4;
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_Q16W16V16U16 SSE2p
void yc_i16_sse2(void *dst, const PIXEL_YC *src, unsigned int n)
{
	__m128i xi0,xi1;
	__m128i *xdst = static_cast<__m128i*>(dst);
	const unsigned int m = n & 1U;
	n /= 2;

	for(unsigned int i=0; i<n; ++i){
		xi0 = _mm_loadl_epi64((const __m128i*)src++);
		xi1 = _mm_loadl_epi64((const __m128i*)src++);
		xi0 = _mm_unpacklo_epi64(xi0, xi1);
		_mm_storeu_si128(xdst++, _mm_and_si128(xi0, i16_msk2.xi));
	}
	if(m){
		xi0 = _mm_loadl_epi64((const __m128i*)src);
		_mm_storel_epi64(xdst, _mm_and_si128(xi0, i16_msk2.xi));
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_A16B16G16R16 SSE2p
void yc_b16_sse2(void *dst, const PIXEL_YC *src, unsigned int n)
{
	__m128i xi0,xi1;
	__m128i *xdst = static_cast<__m128i*>(dst);
	const unsigned int m = n & 1U;
	n /= 2;

	for(unsigned int i=0; i<n; ++i){
		xi0 = _mm_loadl_epi64((const __m128i*)src++);
		xi1 = _mm_loadl_epi64((const __m128i*)src++);
		xi0 = _mm_unpacklo_epi64(xi0 ,xi1);
		xi0 = _mm_add_epi16(xi0, b16_ofs2.xi);
		_mm_storeu_si128(xdst++,  _mm_and_si128(xi0, i16_msk2.xi));
	}
	if(m){
		xi0 = _mm_loadl_epi64((const __m128i*)src);
		xi0 = _mm_add_epi16(xi0, b16_ofs2.xi);
		_mm_storel_epi64(xdst, _mm_and_si128(xi0, i16_msk2.xi));
	}
}

//ϊRs[ D3DFMT_A32B32G32R32F -> PIXEL_YC SSE2p
void f32_yc_sse2(PIXEL_YC *dst, const void *src, unsigned int n)
{
	__m128i xi;
	const float *fsrc = static_cast<const float*>(src);

	for(unsigned int i=0; i<n; ++i){
		xi = _mm_cvtps_epi32(_mm_mul_ps(_mm_loadu_ps(fsrc), f32_rcl));
		_mm_storel_epi64((__m128i*)dst++, _mm_packs_epi32(xi, xi));
		fsrc += 4;
	}
}

//ϊRs[ D3DFMT_A16B16G16R16F -> PIXEL_YC SSE2p
void f16_yc_sse2(PIXEL_YC *dst, const void *src, unsigned int n)
{
	__m128i xi,xs,xz;
	const short *ssrc = static_cast<const short*>(src);

	for(unsigned int i=0; i<n; ++i){
		xi = _mm_loadl_epi64((__m128i*)ssrc);
		xi = _mm_unpacklo_epi16(_mm_setzero_si128(), xi);
		xs = _mm_and_si128(xi, f16_sgn.xi);
		xi = _mm_andnot_si128(f16_sgn.xi, xi);
		xz = _mm_cmpeq_epi32(_mm_setzero_si128(), xi);
		xi = _mm_srli_epi32(xi, 3);
		xi = _mm_add_epi32(xi, f16_rcl3.xi);
		xi = _mm_andnot_si128(xz, _mm_or_si128(xi, xs));
		xi = _mm_cvtps_epi32(_mm_castsi128_ps(xi));
		_mm_storel_epi64((__m128i*)dst++, _mm_packs_epi32(xi, xi));
		ssrc += 4;
	}
}

//ϊRs[ D3DFMT_Q16W16V16U16 -> PIXEL_YC SSE2p
void i16_yc_sse2(PIXEL_YC *dst, const void *src, unsigned int n)
{
	__m128i xi0,xi1;
	const __m128i *xsrc = static_cast<const __m128i*>(src);
	const unsigned int m = n & 1U;
	n /= 2;

	for(unsigned int i=0; i<n; ++i){
		xi0 = _mm_loadu_si128(xsrc++);
		xi1 = _mm_unpackhi_epi64(xi0, xi0);
		_mm_storel_epi64((__m128i*)dst++, xi0);
		_mm_storel_epi64((__m128i*)dst++, xi1);
	}
	if(m){
		_mm_storel_epi64((__m128i*)dst, _mm_loadl_epi64(xsrc));
	}
}

//ϊRs[ D3DFMT_A16B16G16R16 -> PIXEL_YC SSE2p
void b16_yc_sse2(PIXEL_YC *dst, const void *src, unsigned int n)
{
	__m128i xi0,xi1;
	const __m128i *xsrc = static_cast<const __m128i*>(src);
	const unsigned int m = n & 1U;
	n /= 2;

	for(unsigned int i=0; i<n; ++i){
		xi0 = _mm_sub_epi16(_mm_loadu_si128(xsrc++), b16_ofs2.xi);
		xi1 = _mm_unpackhi_epi64(xi0, xi0);
		_mm_storel_epi64((__m128i*)dst++, xi0);
		_mm_storel_epi64((__m128i*)dst++, xi1);
	}
	if(m){
		xi0 = _mm_loadl_epi64(xsrc);
		_mm_storel_epi64((__m128i*)dst, _mm_sub_epi16(xi0, b16_ofs2.xi));
	}
}

//---------------------------------------------------------------------
// Rs[p֐i}Nj
//---------------------------------------------------------------------
#define CPCVT_YC2XX_MT(name) \
static void cc_mt_##name(int, int, void *p, void*)\
{\
	const CpCvtParam *para = static_cast<CpCvtParam*>(p);\
	const PIXEL_YC *src = para->pyc;\
	unsigned char *dst = static_cast<unsigned char*>(para->buf);\
	unsigned int c;\
\
	while((c = _InterlockedExchangeAdd(&CpCvt::s_cmt, 1)) < para->h){\
		##name(dst + c * para->sbuf, src + c * para->spyc, para->w);\
	}

#define CPCVT_XX2YC_MT(name) \
static void cc_mt_##name(int, int, void *p, void*)\
{\
	const CpCvtParam *para = static_cast<CpCvtParam*>(p);\
	const unsigned char *src = static_cast<unsigned char*>(para->buf);\
	PIXEL_YC *dst = para->pyc;\
	unsigned int c;\
\
	while((c = _InterlockedExchangeAdd(&CpCvt::s_cmt, 1)) < para->h){\
		##name(dst + c * para->spyc, src + c * para->sbuf, para->w);\
	}

#define CPCVT_YC2XX(name) \
static void cc_##name(const CpCvtParam &p)\
{\
	const PIXEL_YC *src = p.pyc;\
	unsigned char *dst = static_cast<unsigned char*>(p.buf);\
\
	for(unsigned int i=0; i<p.h; ++i){\
		##name(dst, src, p.w);\
		dst += p.sbuf;\
		src += p.spyc;\
	}

#define CPCVT_XX2YC(name) \
static void cc_##name(const CpCvtParam &p)\
{\
	const unsigned char *src = static_cast<unsigned char*>(p.buf);\
	PIXEL_YC *dst = p.pyc;\
\
	for(unsigned int i=0; i<p.h; ++i){\
		##name(dst, src, p.w);\
		dst += p.spyc;\
		src += p.sbuf;\
	}

CPCVT_YC2XX_MT(yc_f32) }
CPCVT_YC2XX_MT(yc_f32_sse) _mm_empty(); }
CPCVT_YC2XX_MT(yc_f32_sse2) }
CPCVT_YC2XX_MT(yc_f16) }
CPCVT_YC2XX_MT(yc_f16_sse) _mm_sfence(); _mm_empty(); }
CPCVT_YC2XX_MT(yc_f16_sse2) }
CPCVT_YC2XX_MT(yc_i16_mmx) _mm_empty(); }
CPCVT_YC2XX_MT(yc_i16_sse2) }
CPCVT_YC2XX_MT(yc_b16_mmx) _mm_empty(); }
CPCVT_YC2XX_MT(yc_b16_sse2) }

CPCVT_YC2XX(yc_f32) }
CPCVT_YC2XX(yc_f32_sse) _mm_empty(); }
CPCVT_YC2XX(yc_f32_sse2) }
CPCVT_YC2XX(yc_f16) }
CPCVT_YC2XX(yc_f16_sse) _mm_sfence(); _mm_empty(); }
CPCVT_YC2XX(yc_f16_sse2) }
CPCVT_YC2XX(yc_i16_mmx) _mm_empty(); }
CPCVT_YC2XX(yc_i16_sse2) }
CPCVT_YC2XX(yc_b16_mmx) _mm_empty(); }
CPCVT_YC2XX(yc_b16_sse2) }

CPCVT_XX2YC_MT(f32_yc) }
CPCVT_XX2YC_MT(f32_yc_sse) _mm_empty(); }
CPCVT_XX2YC_MT(f32_yc_sse2) }
CPCVT_XX2YC_MT(f16_yc) }
CPCVT_XX2YC_MT(f16_yc_sse) _mm_empty(); }
CPCVT_XX2YC_MT(f16_yc_sse2) }
CPCVT_XX2YC_MT(i16_yc_mmx) _mm_empty(); }
CPCVT_XX2YC_MT(i16_yc_sse2) }
CPCVT_XX2YC_MT(b16_yc_mmx) _mm_empty(); }
CPCVT_XX2YC_MT(b16_yc_sse2) }

CPCVT_XX2YC(f32_yc) }
CPCVT_XX2YC(f32_yc_sse) _mm_empty(); }
CPCVT_XX2YC(f32_yc_sse2) }
CPCVT_XX2YC(f16_yc) }
CPCVT_XX2YC(f16_yc_sse) _mm_empty(); }
CPCVT_XX2YC(f16_yc_sse2) }
CPCVT_XX2YC(i16_yc_mmx) _mm_empty(); }
CPCVT_XX2YC(i16_yc_sse2) }
CPCVT_XX2YC(b16_yc_mmx) _mm_empty(); }
CPCVT_XX2YC(b16_yc_sse2) }


//SIMDΉeXg
static int check_simd()
{
	int cpu_info[4];
	int ret = 0;

	// CPUIDɂ`FbN
	__cpuid(cpu_info, 1);

	// MMX  0x0800000 edx : cpu_info[3]
	// SSE  0x2000000 edx : cpu_info[3]
	// SSE2 0x4000000 edx : cpu_info[3]
	// SSE3 0x0000001 ecx : cpu_info[2]
	if(0x0800000 & cpu_info[3]) ret |= SIMD_MMX;
	if(0x2000000 & cpu_info[3]) ret |= SIMD_SSE;
	if(0x4000000 & cpu_info[3]) ret |= SIMD_SSE2;

	if(ret < SIMD_SSE){
		return ret;
	}

	// OST|[gĂ邩
	__try {
		__asm xorps xmm0, xmm0
	}

	__except(EXCEPTION_EXECUTE_HANDLER){
		if (_exception_code()==STATUS_ILLEGAL_INSTRUCTION)
			/* SSE not supported */
			return ret & SIMD_MMX;
	}

	/* SSE are supported by OS */
	return ret;
}

//---------------------------------------------------------------------
// class CpCvt impl
//---------------------------------------------------------------------
volatile long CpCvt::s_cmt = 0;

bool CpCvt::GetCpConverter(CpCvt* &cc, CpCvtFmt fmt)
{
	static int simd = -1;
	static CpCvt ccs[UNKNOWN];
	
	if(simd < 0){
		simd = check_simd();
		if(simd & SIMD_SSE2){
			ccs[FLOAT32] =
				CpCvt(cc_mt_yc_f32_sse2, cc_mt_f32_yc_sse2, cc_yc_f32_sse2, cc_f32_yc_sse2);
			ccs[FLOAT16] =
				CpCvt(cc_mt_yc_f16_sse2, cc_mt_f16_yc_sse2, cc_yc_f16_sse2, cc_f16_yc_sse2);
			ccs[INT16] =
				CpCvt(cc_mt_yc_i16_sse2, cc_mt_i16_yc_sse2, cc_yc_i16_sse2, cc_i16_yc_sse2);
			ccs[BIASED16] =
				CpCvt(cc_mt_yc_b16_sse2, cc_mt_b16_yc_sse2, cc_yc_b16_sse2, cc_b16_yc_sse2);
		}else if(simd & SIMD_SSE){
			ccs[FLOAT32]  = CpCvt(cc_mt_yc_f32_sse, cc_mt_f32_yc_sse, cc_yc_f32_sse, cc_f32_yc_sse);
			ccs[FLOAT16]  = CpCvt(cc_mt_yc_f16_sse, cc_mt_f16_yc_sse, cc_yc_f16_sse, cc_f16_yc_sse);
			ccs[INT16]    = CpCvt(cc_mt_yc_i16_mmx, cc_mt_i16_yc_mmx, cc_yc_i16_mmx, cc_i16_yc_mmx);
			ccs[BIASED16] = CpCvt(cc_mt_yc_b16_mmx, cc_mt_b16_yc_mmx, cc_yc_b16_mmx, cc_b16_yc_mmx);
		}else{ //MMX͍Œłp\Ƃ
			ccs[FLOAT32]  = CpCvt(cc_mt_yc_f32, cc_mt_f32_yc, cc_yc_f32, cc_f32_yc);
			ccs[FLOAT16]  = CpCvt(cc_mt_yc_f16, cc_mt_f16_yc, cc_yc_f16, cc_f16_yc);
			ccs[INT16]    = CpCvt(cc_mt_yc_i16_mmx, cc_mt_i16_yc_mmx, cc_yc_i16_mmx, cc_i16_yc_mmx);
			ccs[BIASED16] = CpCvt(cc_mt_yc_b16_mmx, cc_mt_b16_yc_mmx, cc_yc_b16_mmx, cc_b16_yc_mmx);
		}
	}

	//ŒłMMXKv
	if(fmt < FLOAT32 || fmt >= UNKNOWN || simd == 0){
		return false;
	}

	cc = &ccs[fmt];
	return true;
}

MULTI_THREAD_FUNC CpCvt::GetYCtoX()
{
	_InterlockedExchange(&s_cmt, 0);
	return m_yc2xx_mt;
}

MULTI_THREAD_FUNC CpCvt::GetXtoYC()
{
	_InterlockedExchange(&s_cmt, 0);
	return m_xx2yc_mt;
}

//---------------------------------------------------------------------
// ptH[}XeXg
//---------------------------------------------------------------------
#ifdef PTEST
#include <iostream>
#include <iomanip>
#include <malloc.h>
using namespace std;

int main()
{
	const int h = 1088;
	const int w = 1920;

	cout << "Size: " << w << "x" << h << " pitch: " << w << endl;

	PIXEL_YC *pyc = (PIXEL_YC*)_aligned_malloc(sizeof(PIXEL_YC) * h * w, 16);
	unsigned char *buf = (PBYTE)_aligned_malloc(h * w * 4 * 4, 16); //fp32 4ch

	CpCvtParam para = { buf, pyc, h, w, w * 4 * 4, w };
	CpCvt::CPCVTF y2x[10] = {
		cc_yc_f32, cc_yc_f32_sse, cc_yc_f32_sse2,
		cc_yc_f16, cc_yc_f16_sse, cc_yc_f16_sse2,
		cc_yc_i16_mmx, cc_yc_i16_sse2, cc_yc_b16_mmx, cc_yc_b16_sse2 };
	char* y2x_desc[10] = {
		"YC->F32     ", "YC->F32 SSE ", "YC->F32 SSE2",
		"YC->F16     ", "YC->F16 SSE ", "YC->F16 SSE2",
		"YC->S16 MMX ", "YC->S16 SSE2", "YC->B16 MMX ", "YC->B16 SSE2" };
	double y2x_tick[10];

	CpCvt::CPCVTF x2y[10] = {
		cc_f32_yc, cc_f32_yc_sse, cc_f32_yc_sse2,
		cc_f16_yc, cc_f16_yc_sse, cc_f16_yc_sse2,
		cc_i16_yc_mmx, cc_i16_yc_sse2, cc_b16_yc_mmx, cc_b16_yc_sse2 };
	char* x2y_desc[10] = {
		"F32->YC     ", "F32->YC SSE ", "F32->YC SSE2",
		"F16->YC     ", "F16->YC SSE ", "F16->YC SSE2",
		"I16->YC MMX ", "I16->YC SSE2", "B16->YC MMX ", "B16->YC SSE2" };
	double x2y_tick[10];

	LARGE_INTEGER qs,qe,qf;
	QueryPerformanceFrequency(&qf);
	double rqf = 1.0/static_cast<double>(qf.QuadPart);

	for(int i=0; i<10; ++i){
		ZeroMemory(pyc, sizeof(PIXEL_YC) * h * w);
		QueryPerformanceCounter(&qs);
		y2x[i](para);
		QueryPerformanceCounter(&qe);
		y2x_tick[i] = static_cast<double>(qe.QuadPart - qs.QuadPart);
	}
	
	for(int i=0; i<10; ++i){
		ZeroMemory(buf, h * w * 4 * 4);
		QueryPerformanceCounter(&qs);
		x2y[i](para);
		QueryPerformanceCounter(&qe);
		x2y_tick[i] = static_cast<double>(qe.QuadPart - qs.QuadPart);
	}

	cout << setiosflags(ios_base::scientific);
	for(int i=0; i<10; ++i){
		cout << y2x_desc[i] << ": " << y2x_tick[i] * rqf << endl;
	}
	
	for(int i=0; i<10; ++i){
		cout << x2y_desc[i] << ": " << x2y_tick[i] * rqf << endl;
	}

	_aligned_free(buf);
	_aligned_free(pyc);

	return 0;
}
#endif
